class AstSelectorException(Exception):
    pass


class UnableToFindElement(AstSelectorException):
    def __init__(self, query: str) -> None:
        self.query = query
        super().__init__(f"Unable to find element with query '{query}' in specified tree")


class UnableToReferenceQuery(AstSelectorException):
    def __init__(self, query: str) -> None:
        self.query = query
        super().__init__(f"Unable to reference '{query}'")



from __future__ import annotations

import ast
from collections import defaultdict
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Generator, Iterator, List, Optional, Type, Union

from .exceptions import UnableToReferenceQuery


@dataclass
class NavigationReference:
    reference_table: Dict[str, List[ast.AST]] = field(default_factory=lambda: defaultdict(list))

    def _build_reference(self, selector_group: SelectorGroup) -> str:
        if selector_group.is_element_selector:
            return f"${selector_group.query}"

        elif selector_group.is_drill_selector:
            last_ref = list(self.reference_table.keys())[-1]
            return f"${last_ref}.{selector_group.query}"

        raise UnableToReferenceQuery(selector_group.query)

    def get_last_entry(self) -> List[ast.AST]:
        return list(self.reference_table.values())[-1]

    def append(self, selector_group: SelectorGroup, val: ast.AST) -> None:
        ref = self._build_reference(selector_group)
        self.reference_table[ref].append(val)

    def get_ref_idx(self, ref: str) -> Optional[int]:
        k = list(reversed(self.reference_table.keys()))
        if ref in k:
            return k.index(ref)

        return None


@dataclass
class SelectorGroup:
    navigation: NavigationReference
    query: str
    referentiable = True

    @property
    def is_attribute_selector(self) -> bool:
        return self.query.startswith("[")

    @property
    def is_drill_selector(self) -> bool:
        return self.query.startswith(".")

    @property
    def is_reference_selector(self) -> bool:
        return self.query.startswith("$")

    @property
    def is_parent_selector(self) -> bool:
        return self.query == ".."

    @property
    def is_element_selector(self) -> bool:
        return all(
            [
                not self.is_attribute_selector,
                not self.is_drill_selector,
                not self.is_reference_selector,
                not self.is_parent_selector,
            ]
        )

    def to_element_selector(self) -> ElementSelector:
        if self.is_element_selector:
            return ElementSelector(self.navigation, self.query)

        if self.is_parent_selector:
            return ParentSelector(self.navigation, self.query)

        if self.is_drill_selector:
            return DrillSelector(self.navigation, self.query)

        if self.is_reference_selector:
            return ReferenceSelector(self.navigation, self.query)

        raise TypeError()

    def to_attribute_selector(self) -> AttributeSelector:
        return AttributeSelector(self.navigation, self.query)

    def _find_nodes(self, branches: Union[Iterator[ast.AST], ast.AST]) -> Generator[ast.AST, None, None]:
        raise NotImplementedError()

    def find_nodes(self, branches: Union[Iterator[ast.AST], ast.AST]) -> Generator[ast.AST, None, None]:
        for node in self._find_nodes(branches):
            if self.referentiable:
                self.navigation.append(self, node)

            yield node


@dataclass
class ElementSelector(SelectorGroup):
    element_type: Type[ast.AST] = field(init=False)
    attr_selectors: List[AttributeSelector] = field(default_factory=list)

    def __post_init__(self) -> None:
        self.element_type = getattr(ast, self.query)

    def append_attr_selector(self, selector: SelectorGroup) -> None:
        self.attr_selectors.append(selector.to_attribute_selector())

    def _matches(self, node: ast.AST) -> bool:
        return all(attr.matches(node) for attr in self.attr_selectors)

    def _find_nodes(self, branches: Union[Iterator[ast.AST], ast.AST]) -> Generator[ast.AST, None, None]:
        if not isinstance(branches, Iterator):
            branches = iter([branches])

        for tree in branches:
            parent = tree
            for node in ast.walk(tree):
                node.parent = parent  # type: ignore
                parent = node
                if isinstance(node, self.element_type):
                    if self._matches(node):
                        yield node


@dataclass
class DrillSelector(ElementSelector):
    def __post_init__(self) -> None:
        self.element_type = ast.AST
        self.query = self.query.lstrip(".")

    def _find_nodes(self, branches: Union[Iterator[ast.AST], ast.AST]) -> Generator[ast.AST, None, None]:
        if not isinstance(branches, Iterator):
            branches = iter([branches])

        for node in list(branches):
            if drilled := getattr(node, self.query, False):
                if isinstance(drilled, ast.AST):
                    if self._matches(drilled):
                        drilled.parent = node  # type: ignore
                        yield drilled


class AttributeSelectorComparator(str, Enum):
    INSTANCE = " is "
    EQUALS = "="


@dataclass
class AttributeSelector(SelectorGroup):
    attr: str = field(init=False)
    condition: AttributeSelectorComparator = field(init=False)
    val: str = field(init=False)

    def __post_init__(self) -> None:
        self.query = self.query.lstrip("[").rstrip("]")
        if AttributeSelectorComparator.EQUALS in self.query:
            self.condition = AttributeSelectorComparator.EQUALS
        else:
            self.condition = AttributeSelectorComparator.INSTANCE

        self.attr, self.val = [x.strip() for x in self.query.split(self.condition)]

    def _match_equals(self, node: ast.AST) -> bool:
        attr_val = str(getattr(node, self.attr))
        expected_val = self.val

        return attr_val == expected_val

    def _match_instance(self, node: ast.AST) -> bool:
        attr_val = getattr(node, self.attr)
        expected_val_type = getattr(ast, self.val, type(None))

        if isinstance(attr_val, expected_val_type):
            return True

        return False

    def matches(self, node: ast.AST) -> bool:
        if self.condition == AttributeSelectorComparator.INSTANCE:
            return self._match_instance(node)

        return self._match_equals(node)


@dataclass
class ReferenceSelector(ElementSelector):
    referentiable = False

    def __post_init__(self) -> None:
        self.element_type = ast.AST

    def _find_nodes(self, branches: Union[Iterator[ast.AST], ast.AST]) -> Generator[ast.AST, None, None]:
        if not isinstance(branches, Iterator):
            branches = iter([branches])

        tree = list(branches)  # Force generators to build reference table until this point
        parent_level = self.navigation.get_ref_idx(self.query)
        if parent_level:
            for node in tree:
                selected = node
                for _ in range(parent_level):
                    selected = selected.parent  # type: ignore

                if self._matches(selected):
                    yield selected


@dataclass
class ParentSelector(ReferenceSelector):
    def _find_nodes(self, branches: Union[Iterator[ast.AST], ast.AST]) -> Generator[ast.AST, None, None]:
        if not isinstance(branches, Iterator):
            branches = iter([branches])

        for node in branches:
            yield node.parent  # type: ignore



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['ast_selector']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'ast-selector',
    'version': '0.1.4',
    'description': 'Query AST elements by using CSS Selector-like syntax',
    'long_description': '# AST Selector\n\n<p align="center">\n    <img src="https://raw.githubusercontent.com/guilatrova/ast_selector/main/img/logo.png">\n</p>\n\n<h2 align="center">Query AST elements by using CSS Selector-like syntax</h2>\n\n<p align="center">\n  <a href="https://github.com/guilatrova/ast_selector/actions"><img alt="Actions Status" src="https://github.com/guilatrova/ast_selector/workflows/CI/badge.svg"></a>\n  <a href="https://pypi.org/project/ast-selector/"><img alt="PyPI" src="https://img.shields.io/pypi/v/ast_selector"/></a>\n  <img src="https://badgen.net/pypi/python/ast_selector" />\n  <a href="https://github.com/relekang/python-semantic-release"><img alt="Semantic Release" src="https://img.shields.io/badge/%20%20%F0%9F%93%A6%F0%9F%9A%80-semantic--release-e10079.svg"></a>\n  <a href="https://github.com/guilatrova/ast_selector/blob/main/LICENSE"><img alt="GitHub" src="https://img.shields.io/github/license/guilatrova/ast_selector"/></a>\n  <a href="https://pepy.tech/project/ast-selector/"><img alt="Downloads" src="https://static.pepy.tech/personalized-badge/ast_selector?period=total&units=international_system&left_color=grey&right_color=blue&left_text=%F0%9F%A6%96%20Downloads"/></a>\n  <a href="https://github.com/psf/black"><img alt="Code style: black" src="https://img.shields.io/badge/code%20style-black-000000.svg"/></a>\n  <a href="https://github.com/guilatrova/tryceratops"><img alt="try/except style: tryceratops" src="https://img.shields.io/badge/try%2Fexcept%20style-tryceratops%20%F0%9F%A6%96%E2%9C%A8-black" /></a>\n  <a href="https://twitter.com/intent/user?screen_name=guilatrova"><img alt="Follow guilatrova" src="https://img.shields.io/twitter/follow/guilatrova?style=social"/></a>\n</p>\n\n> "Query AST elements 🌲 by using CSS Selector-like 💅 syntax."\n\n## Installation and usage\n\n### Installation\n\n```\npip install ast_selector\n```\n\n### Usage\n\n```py\nfrom ast_selector import AstSelector\n\ntree = load_python_code_as_ast_tree()\nquery = "FunctionDef Raise $FunctionDef"\n# Query all functions that raises at least an exception\n\nfunctions_raising_exceptions = AstSelector(query, tree).all()\n```\n\n### Use Cases\n\n#### Functions that return int\n\n```py\nfrom ast_selector import AstSelector\n\ntree = load_python_code_as_ast_tree()\nquery = "FunctionDef.returns[id=int] $FunctionDef"\n# Query all functions that return ints e.g. def sum() -> int\n\nfunction_element = AstSelector(query, tree).first()\n```\n\n## License\n\nMIT\n\n## Credits\n\nIt\'s extremely hard to keep hacking on open source like that while keeping a full-time job. I thank God from the bottom of my heart for both the inspiration and the energy.\n',
    'author': 'Guilherme Latrova',
    'author_email': 'hello@guilatrova.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/guilatrova/ast_selector',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import ast
import itertools
import re
from typing import Generator, List, Optional

from .exceptions import UnableToFindElement
from .models import ElementSelector, NavigationReference, SelectorGroup


class AstSelector:
    """
    CSS(Ast) Selector:

    Raise = instance type
    > [el] = direct child
    .abc = ????
    #abc = ????
    [attr is X] = any element where attr is X
    [attr is None] = any element where attr is X
    [a=c][b=d] = combine attr selectors
    """

    def __init__(self, query: str, tree: ast.AST) -> None:
        self.tree = tree
        self.query = query
        # TODO: Validate query

    def _resolve_query(self) -> List[ElementSelector]:
        PARENT_RE = r"(\.\.)?"
        DRILL_RE = r"(\.\w+)?"
        ATTR_RE = r"(\[[a-zA-Z0-9_= ]+\])?"
        NODE_RE = r"(\$?[A-Z]\w+)?"

        reg = re.findall(f"{PARENT_RE}{DRILL_RE}{ATTR_RE}{NODE_RE}", self.query)
        reggroup = [x for x in itertools.chain(*reg) if x]

        el_selector: Optional[ElementSelector] = None
        reference_table = NavigationReference()
        results = []

        for g in reggroup:
            selector = SelectorGroup(reference_table, g)

            if selector.is_attribute_selector:
                if el_selector is not None:
                    el_selector.append_attr_selector(selector)
            else:
                el_selector = selector.to_element_selector()
                results.append(el_selector)

        return results

    def _resolve(self) -> Generator[ast.AST, None, None]:
        groups = self._resolve_query()

        tree = iter([self.tree])
        for group in groups:
            tree = group.find_nodes(tree)

        yield from tree

    def exists(self) -> bool:
        el = self._resolve()
        try:
            next(el)
        except StopIteration:
            return False
        else:
            return True

    def count(self) -> int:
        nodes = self._resolve()
        return len(list(nodes))

    def all(self) -> List[ast.AST]:
        nodes = self._resolve()
        return list(nodes)

    def first(self) -> ast.AST:
        el = self._resolve()
        try:
            first = next(el)
        except StopIteration as e:
            raise UnableToFindElement(self.query) from e
        else:
            return first
